#include <cassert>
#include <vector>
#include <utility>
#include <iostream>
#include <cstring>

#include <unistd.h>
#include <sys/ioctl.h>

// signal handling (ctrl-c)
#include <signal.h>

// libdrm
#include <xf86drm.h>
#include <xf86drmMode.h>
// intel-specific memory management
#include <libdrm/intel_bufmgr.h>
#include <libdrm/i915_drm.h>
// intel-specific OPCODES supported by the GPU the information here is mainly
// from the mesa i915 driver:
//
// https://cgit.freedesktop.org/mesa/mesa/tree/src/mesa/drivers/dri/i915/intel_reg.h
// "Type" of the opcode
#define CMD_MI              (0x0 << 29)
#define CMD_2D              (0x2 << 29)
// Commands
#define XY_COLOR_BLT_CMD    (CMD_2D | (0x50 << 22))
// Which channels from rgba to write
#define XY_BLT_WRITE_ALPHA  (1 << 21)
#define XY_BLT_WRITE_RGB    (1 << 20)
// Terminate a commands buffer
#define MI_BATCH_BUFFER_END (CMD_MI | 0xA << 23)

// Global state to keep track of when to exit
bool received_signint = false;

// Put everything needed to draw to a framebuffer (i.e.: the screen) in one
// place so we can easily pass everything around.
struct intel_drawing_context {
  // Keeps track of where to send the ioctl() calls to
  const int gpu;
  // Keeps state for the vram memory allocator
  drm_intel_bufmgr *bufmgr;
  // Keeps track of the current "process" inside the graphics card to execute
  // commands
  drm_intel_context *gpu_ctx;
  // This is the actual buffer containing the pixels in the screen
  drm_intel_bo *buffer;
  // The pitch for the screen buffer (i.e. width * bytes per pixel)
  const uint32_t pitch;
};

// Easy interface to send a sequence of commands to the BLT engine of the
// graphics card. Intel has multiple engines inside its cards: video, 2d (BLT),
// 3d.
//
// This also takes care of relocating any memory addresses of other buffer
// objects referenced in the commands to the correct vram address, with the
// information given in `refs`, which is a mapping between an index in this
// buffer and the buffer that index should point to.
void send_blt_commands(struct intel_drawing_context& ctx,
                       std::vector<std::pair<size_t, drm_intel_bo*>> refs,
                       std::vector<uint32_t> commands) {
  // Convert the number of commands to the length of the command buffer in bytes
  const auto length = commands.size()*sizeof(uint32_t);

  // Create the buffer
  auto cmds_bo = drm_intel_bo_alloc(ctx.bufmgr, "commands", length, 0);

  // Map to memory, 1 = writable
  // and write the commands to it
  drm_intel_bo_map(cmds_bo, 1);
  memcpy(cmds_bo->virt, &commands.front(), length);
  drm_intel_bo_unmap(cmds_bo);

  // Setup all addresses referenced
  for (auto ref: refs) {
    // For each address update the original command buffer with the actual
    // memory address of the referenced buffer object
    assert(drm_intel_bo_emit_reloc(cmds_bo, ref.first*4,
                                   ref.second, 0,
                                   I915_GEM_DOMAIN_RENDER, I915_GEM_DOMAIN_RENDER) == 0);
  }

  assert(drm_intel_gem_bo_context_exec(cmds_bo, ctx.gpu_ctx, length, I915_EXEC_BLT) == 0);

  drm_intel_bo_unreference(cmds_bo);
}

void intel_2d_rect(struct intel_drawing_context& ctx,
                   uint32_t x, uint32_t y, uint32_t width, uint32_t height,
                   uint32_t color) {
  send_blt_commands(ctx, {{4, ctx.buffer}}, {
    XY_COLOR_BLT_CMD | XY_BLT_WRITE_ALPHA | XY_BLT_WRITE_RGB | 0x5,
    (0 << 30) | (0b11 << 24) | (0xf0 << 16) | ctx.pitch,
    (y << 16) | x,
    ((y+height) << 16) | (x+width),
    0,
    0,
    color,
    MI_BATCH_BUFFER_END
  });
}

void on_sigint(int) {
  std::cout << "Will exit in the next iteration." << std::endl;

  received_signint = true;
}

void draw_loop(const int gpu,
               drm_intel_bufmgr* bufmgr,
               drm_intel_bo *screen,
               uint32_t width, uint32_t height) {
  // Create a context that will be used to execute commands
  auto context = drm_intel_gem_context_create(bufmgr);

  // Build a drawing context to draw things on
  struct intel_drawing_context ctx = {
    .gpu = gpu,
    .bufmgr = bufmgr,
    .gpu_ctx = context,
    .buffer = screen,
    .pitch = width*4+32
  };

  while (!received_signint) {
    const uint32_t x = rand()%width, y = rand()%height;
    const uint32_t w = rand()%width, h = rand()%height;

    intel_2d_rect(ctx, x, y, w, h, rand());
  }

  drm_intel_gem_context_destroy(context);
}

void register_ctrl_c_handler() {
  struct sigaction sigint;

  sigint.sa_handler = on_sigint;
  sigint.sa_flags = 0;

  sigaction(SIGINT, &sigint, &sigint);
}

// To debug DRM calls we can run the program with LIBGL_DEBUG=verbose
int main(int arg_count, char** arg_vector) {
  register_ctrl_c_handler();

  assert(drmAvailable());

  // There is a drmOpenMinor function in libdrm for which no Bus ID is needed
  // but it is not exposed :/
  // We could also directly open the file and use that without any issue
  const int gpu = drmOpen(nullptr, "pci:0000:00:02.0");
  auto bufmgr = drm_intel_bufmgr_gem_init(gpu, 4096);

  // This can be used to debug memory management calls:
  // drm_intel_bufmgr_set_debug(bufmgr, 1);

  assert(gpu >= 0);

  const drmModeResPtr res = drmModeGetResources(gpu);

  std::cout << "Card supports from "
            << res->min_width << "x" << res->min_height
            << " to "
            << res->max_width << "x" << res->max_height
            << std::endl;

  for (size_t c=0;c<res->count_connectors;c++) {
    std::cout << "Connector " << c << std::endl;

    auto connector = drmModeGetConnector(gpu, res->connectors[c]);

    std::cout << connector->mmWidth << "x" << connector->mmHeight
              << "mm" << std::endl;

    for (size_t m=0;m<connector->count_modes;m++) {
      std::cout << "Mode " << m << std::endl;

      std::cout << connector->modes[m].name << std::endl;

      auto width = connector->modes[m].hdisplay;
      auto height = connector->modes[m].vdisplay;

      std::cout << width << "x" << height << std::endl;

      auto encoder = drmModeGetEncoder(gpu, connector->encoder_id);
      auto controller = drmModeGetCrtc(gpu, encoder->crtc_id);

      uint32_t connector_ids[] = {
        connector->connector_id
      };

      // "Buffer Object"
      uint32_t tiling_mode = I915_TILING_NONE;
      uint64_t pitch = 4*width;
      auto screen = drm_intel_bo_alloc_tiled(bufmgr, "frame",
                                             width, height, 4,
                                             &tiling_mode, &pitch,
                                             0);
      uint32_t buf_id = 0;

      assert(drmModeAddFB(gpu, width, height, 24, 32, pitch, screen->handle, &buf_id) == 0);

      assert(drmModeSetCrtc(gpu, controller->crtc_id,
                            buf_id,
                            0, 0,
                            connector_ids, 1,
                            &connector->modes[m]) == 0);

      draw_loop(gpu, bufmgr, screen, width, height);

      drm_intel_bo_unreference(screen);

      drmModeFreeCrtc(controller);
      drmModeFreeEncoder(encoder);
      break;
    }

    drmModeFreeConnector(connector);
    break;
  }

  drmModeFreeResources(res);

  drm_intel_bufmgr_destroy(bufmgr);

  assert(drmClose(gpu) == 0);

  return 0;
}
